@@ -39,8 +39,11 @@
 #include "nautilus-desktop-window.h"
 #include "nautilus-first-time-druid.h"
 #include "nautilus-main.h"
+#include "nautilus-spatial-window.h"
+#include "nautilus-navigation-window.h"
 #include "nautilus-shell-interface.h"
 #include "nautilus-shell.h"
+#include "nautilus-window-private.h"
 #include <bonobo/bonobo-main.h>
 #include <bonobo/bonobo-object.h>
 #include <eel/eel-gtk-macros.h>
@@ -62,6 +65,7 @@
 #include <libgnomevfs/gnome-vfs-mime-handlers.h>
 #include <libgnomevfs/gnome-vfs-ops.h>
 #include <libgnomevfs/gnome-vfs-utils.h>
+#include <libgnomevfs/gnome-vfs-volume-monitor.h>
 #include <libnautilus-private/nautilus-file-utilities.h>
 #include <libnautilus-private/nautilus-global-preferences.h>
 #include <libnautilus-private/nautilus-icon-factory.h>
@@ -69,7 +73,6 @@
 #include <libnautilus-private/nautilus-sound.h>
 #include <libnautilus-private/nautilus-bonobo-extensions.h>
 #include <libnautilus-private/nautilus-undo-manager.h>
-#include <libnautilus-private/nautilus-volume-monitor.h>
 #include <libnautilus-private/nautilus-desktop-link-monitor.h>
 #include <libnautilus-private/nautilus-directory-private.h>
 #include <bonobo-activation/bonobo-activation.h>
@@ -88,11 +91,17 @@ static GList *nautilus_application_desktop_windows;
 /* Keeps track of all the nautilus windows. */
 static GList *nautilus_application_window_list;
 
+/* Keeps track of all the object windows */
+static GList *nautilus_application_spatial_window_list;
+
 static gboolean need_to_show_first_time_druid     (void);
 static void     desktop_changed_callback          (gpointer                  user_data);
 static void     desktop_location_changed_callback (gpointer                  user_data);
-static void     volume_unmounted_callback         (NautilusVolumeMonitor    *monitor,
-						   NautilusVolume           *volume,
+static void     volume_unmounted_callback         (GnomeVFSVolumeMonitor    *monitor,
+						   GnomeVFSVolume           *volume,
+						   NautilusApplication      *application);
+static void     volume_mounted_callback           (GnomeVFSVolumeMonitor    *monitor,
+						   GnomeVFSVolume           *volume,
 						   NautilusApplication      *application);
 static void     update_session                    (gpointer                  callback_data);
 static void     init_session                      (void);
@@ -140,6 +149,19 @@ nautilus_application_get_window_list (void)
 	return nautilus_application_window_list;
 }
 
+GList *
+nautilus_application_get_spatial_window_list (void)
+{
+	return nautilus_application_spatial_window_list;
+}
+
+static CORBA_Object
+create_object_shortcut (const char *iid,
+			gpointer callback_data)
+{
+	return create_object (BONOBO_OBJREF (callback_data), iid, NULL);
+}
+
 static void
 nautilus_application_instance_init (NautilusApplication *application)
 {
@@ -151,8 +173,15 @@ nautilus_application_instance_init (NautilusApplication *application)
 	 * used anymore */
 
 	/* Watch for volume unmounts so we can close open windows */
-	g_signal_connect_object (nautilus_volume_monitor_get (), "volume_unmounted",
+	g_signal_connect_object (gnome_vfs_get_volume_monitor (), "volume_unmounted",
 				 G_CALLBACK (volume_unmounted_callback), application, 0);
+	g_signal_connect_object (gnome_vfs_get_volume_monitor (), "volume_mounted",
+				 G_CALLBACK (volume_mounted_callback), application, 0);
+
+	nautilus_bonobo_register_activation_shortcut (NAUTILUS_ICON_VIEW_IID, create_object_shortcut, application);
+	nautilus_bonobo_register_activation_shortcut (NAUTILUS_DESKTOP_ICON_VIEW_IID, create_object_shortcut, application);
+	nautilus_bonobo_register_activation_shortcut (NAUTILUS_LIST_VIEW_IID, create_object_shortcut, application);
+	nautilus_bonobo_register_activation_shortcut (SEARCH_LIST_VIEW_IID, create_object_shortcut, application);
 }
 
 NautilusApplication *
@@ -176,6 +205,11 @@ nautilus_application_destroy (BonoboObject *object)
 
 	application = NAUTILUS_APPLICATION (object);
 
+	nautilus_bonobo_unregister_activation_shortcut (NAUTILUS_ICON_VIEW_IID);
+	nautilus_bonobo_unregister_activation_shortcut (NAUTILUS_DESKTOP_ICON_VIEW_IID);
+	nautilus_bonobo_unregister_activation_shortcut (NAUTILUS_LIST_VIEW_IID);
+	nautilus_bonobo_unregister_activation_shortcut (SEARCH_LIST_VIEW_IID);
+	
 	nautilus_bookmarks_exiting ();
 	
 	bonobo_object_unref (application->undo_manager);
@@ -423,6 +457,7 @@ nautilus_application_startup (NautilusApplication *application,
 			      gboolean no_default_window,
 			      gboolean no_desktop,
 			      gboolean do_first_time_druid_check,
+			      gboolean browser_window,
 			      const char *geometry,
 			      const char *urls[])
 {
@@ -596,10 +631,10 @@ nautilus_application_startup (NautilusApplication *application,
 	  	/* Create the other windows. */
 		if (urls != NULL) {
 			url_list = nautilus_make_uri_list_from_shell_strv (urls);
-			Nautilus_Shell_open_windows (shell, url_list, corba_geometry, &ev);
+			Nautilus_Shell_open_windows (shell, url_list, corba_geometry, browser_window, &ev);
 			CORBA_free (url_list);
 		} else if (!no_default_window) {
-			Nautilus_Shell_open_default_window (shell, corba_geometry, &ev);
+			Nautilus_Shell_open_default_window (shell, corba_geometry, browser_window, &ev);
 		}
 		
 		/* Add ourselves to the session */
@@ -756,10 +791,108 @@ nautilus_application_close_desktop (void)
 }
 
 void
-nautilus_application_close_all_windows (void)
+nautilus_application_close_all_navigation_windows (void)
 {
-	while (nautilus_application_window_list != NULL) {
-		nautilus_window_close (NAUTILUS_WINDOW (nautilus_application_window_list->data));
+	GList *list_copy;
+	GList *l;
+	
+	list_copy = g_list_copy (nautilus_application_window_list);
+	for (l = list_copy; l != NULL; l = l->next) {
+		NautilusWindow *window;
+		
+		window = NAUTILUS_WINDOW (l->data);
+		
+		if (NAUTILUS_IS_NAVIGATION_WINDOW (window)) {
+			nautilus_window_close (window);
+		}
+	}
+	g_list_free (list_copy);
+}
+
+static NautilusSpatialWindow *
+nautilus_application_get_existing_spatial_window (const char *location)
+{
+	GList *l;
+	
+	for (l = nautilus_application_get_spatial_window_list ();
+	     l != NULL; l = l->next) {
+		char *window_location;
+		
+		window_location = nautilus_window_get_location (NAUTILUS_WINDOW (l->data));
+		if (!strcmp (location, window_location)) {
+			g_free (window_location);
+			return NAUTILUS_SPATIAL_WINDOW (l->data);
+		}
+		g_free (window_location);
+	}
+	return NULL;
+}
+
+static NautilusSpatialWindow *
+find_parent_spatial_window (NautilusSpatialWindow *window)
+{
+	NautilusFile *file;
+	NautilusFile *parent_file;
+	char *location;
+	char *desktop_directory;
+
+	location = nautilus_window_get_location (NAUTILUS_WINDOW (window));
+	file = nautilus_file_get (location);
+	g_free (location);
+
+	if (!file) {
+		return NULL;
+	}
+	
+	desktop_directory = nautilus_get_desktop_directory_uri ();	
+	
+	parent_file = nautilus_file_get_parent (file);
+	nautilus_file_unref (file);
+	while (parent_file) {
+		NautilusSpatialWindow *parent_window;
+		
+		location = nautilus_file_get_uri (parent_file);
+
+		/* Stop at the desktop directory, as this is the
+		 * conceptual root of the spatial windows */
+		if (!strcmp (location, desktop_directory)) {
+			g_free (location);
+			g_free (desktop_directory);
+			nautilus_file_unref (parent_file);
+			return NULL;
+		}
+
+		parent_window = nautilus_application_get_existing_spatial_window (location);
+		g_free (location);
+		
+		if (parent_window) {
+			nautilus_file_unref (parent_file);
+			return parent_window;
+		}
+		file = parent_file;
+		parent_file = nautilus_file_get_parent (file);
+		nautilus_file_unref (file);
+	}
+	g_free (desktop_directory);
+
+	return NULL;
+}
+
+void
+nautilus_application_close_with_parent_windows (NautilusSpatialWindow *window)
+{
+	NautilusSpatialWindow *parent_window;
+
+	g_return_if_fail (NAUTILUS_IS_SPATIAL_WINDOW (window));
+
+	parent_window = find_parent_spatial_window (window);
+	nautilus_window_close (NAUTILUS_WINDOW (window));
+	window = parent_window;
+	
+	while (parent_window) {
+		parent_window = find_parent_spatial_window (window);
+		nautilus_window_close (NAUTILUS_WINDOW (window));
+		window = parent_window;
 	}
 }
 
@@ -782,108 +915,29 @@ nautilus_window_delete_event_callback (GtkWidget *widget,
 	return TRUE;
 }				       
 
-static gboolean
-save_window_geometry_timeout (gpointer callback_data)
+
+static NautilusWindow *
+create_window (NautilusApplication *application,
+	       GType window_type,
+	       GdkScreen *screen)
 {
 	NautilusWindow *window;
 	
-	window = NAUTILUS_WINDOW (callback_data);
-	
-	nautilus_window_save_geometry (window);
-
-	window->save_geometry_timeout_id = 0;
-	return FALSE;
-}
- 
-static gboolean
-nautilus_window_configure_event_callback (GtkWidget *widget,
-						GdkEventConfigure *event,
-						gpointer callback_data)
-{
-	NautilusWindow *window;
-	char *geometry_string;
-	
-	window = NAUTILUS_WINDOW (widget);
-	
-	/* Only save the geometry if the user hasn't resized the window
-	 * for half a second. Otherwise delay the callback another half second.
-	 */
-	if (window->save_geometry_timeout_id != 0) {
-		g_source_remove (window->save_geometry_timeout_id);	
-	}
-	if (GTK_WIDGET_VISIBLE (GTK_WIDGET (window)) && !NAUTILUS_IS_DESKTOP_WINDOW (window)) {
-		
-		geometry_string = eel_gtk_window_get_geometry_string (GTK_WINDOW (window));
-	
-		/* If the last geometry is NULL the window must have just
-		 * been shown. No need to save geometry to disk since it
-		 * must be the same.
-		 */
-		if (window->last_geometry == NULL) {
-			window->last_geometry = geometry_string;
-			return FALSE;
-		}
-	
-		/* Don't save geometry if it's the same as before. */
-		if (!strcmp (window->last_geometry, geometry_string)) {
-			g_free (geometry_string);
-			return FALSE;
-		}
-
-		g_free (window->last_geometry);
-		window->last_geometry = geometry_string;
-
-		window->save_geometry_timeout_id = 
-				g_timeout_add (500, save_window_geometry_timeout, window);
-	}
-
-	return FALSE;
-}
-
-static gboolean
-nautilus_window_unrealize_event_callback (GtkWidget *widget,
-						GdkEvent *event,
-						gpointer callback_data)
-{
-	NautilusWindow *window;
-	
-	window = NAUTILUS_WINDOW (widget);
-
-	if (window->save_geometry_timeout_id != 0) {
-		g_source_remove (window->save_geometry_timeout_id);
-		window->save_geometry_timeout_id = 0;
-		nautilus_window_save_geometry (window);
-	}
-
-	return FALSE;
-}
-
-NautilusWindow *
-nautilus_application_create_window (NautilusApplication *application,
-				    GdkScreen           *screen)
-{
-	NautilusWindow *window;
-
 	g_return_val_if_fail (NAUTILUS_IS_APPLICATION (application), NULL);
 	
-	window = NAUTILUS_WINDOW (gtk_widget_new (nautilus_window_get_type (),
+	window = NAUTILUS_WINDOW (gtk_widget_new (window_type,
 						  "app", application,
 						  "app_id", "nautilus",
 						  "screen", screen,
 						  NULL));
 
-	g_signal_connect (window, "delete_event",
-			  G_CALLBACK (nautilus_window_delete_event_callback), NULL);
+	g_signal_connect_data (window, "delete_event",
+			       G_CALLBACK (nautilus_window_delete_event_callback), NULL, NULL,
+			       G_CONNECT_AFTER);
 
 	g_signal_connect_object (window, "destroy",
 				 G_CALLBACK (nautilus_application_destroyed_window), application, 0);
 
-	g_signal_connect (window, "configure_event",
-			  G_CALLBACK (nautilus_window_configure_event_callback), NULL);
-
-	g_signal_connect (window, "unrealize",
-			  G_CALLBACK (nautilus_window_unrealize_event_callback), NULL);
-
 	nautilus_application_window_list = g_list_prepend (nautilus_application_window_list, window);
 
 	/* Do not yet show the window. It will be shown later on if it can
@@ -894,6 +948,65 @@ nautilus_application_create_window (NautilusApplication *application,
 	return window;
 }
 
+static void
+spatial_window_destroyed_callback (void *user_data, GObject *window)
+{
+	nautilus_application_spatial_window_list = g_list_remove (nautilus_application_spatial_window_list, window);
+		
+}
+
+NautilusWindow *
+nautilus_application_present_spatial_window (NautilusApplication *application,
+					    const char          *location,
+					    GdkScreen           *screen)
+{
+	NautilusWindow *window;
+	GList *l;
+
+	g_return_val_if_fail (NAUTILUS_IS_APPLICATION (application), NULL);
+
+	for (l = nautilus_application_get_spatial_window_list ();
+	     l != NULL; l = l->next) {
+		NautilusWindow *existing_window;
+		char *existing_location;
+               	     
+		existing_window = NAUTILUS_WINDOW (l->data);
+		existing_location = existing_window->details->pending_location;
+                
+		if (existing_location == NULL) {
+			existing_location = existing_window->details->location;
+		}
+
+		if (eel_uris_match (existing_location, location)) {
+			gtk_window_present (GTK_WINDOW (existing_window));
+			g_object_ref (existing_window);
+			return existing_window;
+		}
+	}
+
+	window = create_window (application, NAUTILUS_TYPE_SPATIAL_WINDOW, screen);
+	nautilus_application_spatial_window_list = g_list_prepend (nautilus_application_spatial_window_list, window);
+	g_object_weak_ref (G_OBJECT (window), 
+			   spatial_window_destroyed_callback, NULL);
+	
+	nautilus_window_go_to (window, location);
+	
+	return window;
+}
+
+NautilusWindow *
+nautilus_application_create_navigation_window (NautilusApplication *application,
+					       GdkScreen           *screen)
+{
+	NautilusWindow *window;
+
+	g_return_val_if_fail (NAUTILUS_IS_APPLICATION (application), NULL);
+	
+	window = create_window (application, NAUTILUS_TYPE_NAVIGATION_WINDOW, screen);
+
+	return window;
+}
+
 /* callback for changing the directory the desktop points to */
 static void
 desktop_location_changed_callback (gpointer user_data)
@@ -978,64 +1091,59 @@ window_can_be_closed (NautilusWindow *window)
 	return FALSE;
 }
 
-static gboolean
-is_last_closable_window (NautilusWindow *window)
+static void
+volume_mounted_callback (GnomeVFSVolumeMonitor *monitor,
+			 GnomeVFSVolume *volume,
+			 NautilusApplication *application)
 {
-	GList *node, *window_list;
-	
-	window_list = nautilus_application_get_window_list ();
-	
-	for (node = window_list; node != NULL; node = node->next) {
-		if (window != NAUTILUS_WINDOW (node->data) && window_can_be_closed (NAUTILUS_WINDOW (node->data))) {
-			return FALSE;
-		}
+	char *activation_uri;
+	NautilusDirectory *directory;
+		
+	activation_uri = gnome_vfs_volume_get_activation_uri (volume);
+	directory = nautilus_directory_get_existing (activation_uri);
+	g_free (activation_uri);
+	if (directory != NULL) {
+		nautilus_directory_force_reload (directory);
+		nautilus_directory_unref (directory);
 	}
-	
-	return TRUE;
 }
 
-
 /* Called whenever a volume is unmounted. Check and see if there are any windows open
  * displaying contents on the volume. If there are, close them.
  * It would also be cool to save open window and position info.
  */
 static void
-volume_unmounted_callback (NautilusVolumeMonitor *monitor, NautilusVolume *volume,
+volume_unmounted_callback (GnomeVFSVolumeMonitor *monitor,
+			   GnomeVFSVolume *volume,
 			   NautilusApplication *application)
 {
 	GList *window_list, *node, *close_list;
 	NautilusWindow *window;
-	char *uri;
-	char *path;
+	char *uri, *activation_uri;
 		
 	close_list = NULL;
 	
 	/* Check and see if any of the open windows are displaying contents from the unmounted volume */
 	window_list = nautilus_application_get_window_list ();
-	
+
+	activation_uri = gnome_vfs_volume_get_activation_uri (volume);
 	/* Construct a list of windows to be closed. Do not add the non-closable windows to the list. */
 	for (node = window_list; node != NULL; node = node->next) {
 		window = NAUTILUS_WINDOW (node->data);
 		if (window != NULL && window_can_be_closed (window)) {
 			uri = nautilus_window_get_location (window);
-			path = gnome_vfs_get_local_path_from_uri (uri);
-			if (eel_str_has_prefix (path, nautilus_volume_get_mount_path (volume))) {
+			if (eel_str_has_prefix (uri, activation_uri)) {
 				close_list = g_list_prepend (close_list, window);
 			}
-			g_free (path);
 			g_free (uri);
 		}
 	}
+	g_free (activation_uri);
 		
 	/* Handle the windows in the close list. */
 	for (node = close_list; node != NULL; node = node->next) {
 		window = NAUTILUS_WINDOW (node->data);
-		if (is_last_closable_window (window)) {
-			/* Don't close the last or only window. Try to redirect to the default home directory. */		 	
-			nautilus_window_go_home (window);
-		} else {
-			nautilus_window_close (window);
-		}
+		nautilus_window_close (window);
 	}
 		
 	g_list_free (close_list);
